@@ -10,13 +10,13 @@
 #include <config.h>
 #endif
 
+#include <libgnome/gnome-i18n.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <fcntl.h>
 #include <linux/cdrom.h>
 #include <errno.h>
 
-#include "gnome-cd.h"
 #include "linux-cdrom.h"
 
 static GnomeCDRomClass *parent_class = NULL;
@@ -138,19 +138,25 @@ linux_cdrom_open (LinuxCDRom *lcd,
 
 	lcd->priv->cdrom_fd = open (lcd->priv->cdrom_device, O_RDONLY | O_NONBLOCK);
 	if (lcd->priv->cdrom_fd < 0) {
-		if (error) {
+		if (errno == EACCES && error != NULL) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_NOT_OPENED,
-					      "Unable to open %s. This may be caused by:\n"
-					      "a) CD support is not compiled into Linux\n"
-					      "b) You do not have the correct permissions to access the CD drive\n"
-					      "c) %s is not the CD drive.\n",					     
+					      _("You do not seem to have permission to access %s."),
+					      lcd->priv->cdrom_device);
+		} else if (error != NULL) {
+			*error = g_error_new (GNOME_CDROM_ERROR,
+					      GNOME_CDROM_ERROR_NOT_OPENED,
+					      _("%s does not appear to point to a valid CD device. This may be because:\n"
+						"a) CD support is not compiled into Linux\n"
+						"b) You do not have the correct permissions to access the CD drive\n"
+						"c) %s is not the CD drive.\n"),					     
 					      lcd->priv->cdrom_device, lcd->priv->cdrom_device);
 		}
-
+		
 		return FALSE;
 	}
 
+	
 	lcd->priv->ref_count = 1;
 	return TRUE;
 }
@@ -215,12 +221,12 @@ linux_cdrom_update_cd (LinuxCDRom *lcd)
 	priv = lcd->priv;
 
 	if (linux_cdrom_open (lcd, &error) == FALSE) {
-		g_warning ("Error opening CD");
+		g_warning (_("Error opening CD"));
 		return;
 	}
 
 	if (ioctl (priv->cdrom_fd, CDROMREADTOCHDR, priv->tochdr) < 0) {
-		g_warning ("Error reading CD header");
+		g_warning (_("Error reading CD header"));
 		linux_cdrom_close (lcd);
 
 		return;
@@ -237,7 +243,7 @@ linux_cdrom_update_cd (LinuxCDRom *lcd)
 		tocentry.cdte_format = CDROM_MSF;
 
 		if (ioctl (priv->cdrom_fd, CDROMREADTOCENTRY, &tocentry) < 0) {
-			g_warning ("IOCtl failed");
+			g_warning (_("IOCtl failed"));
 			continue;
 		}
 
@@ -249,7 +255,7 @@ linux_cdrom_update_cd (LinuxCDRom *lcd)
 	tocentry.cdte_track = CDROM_LEADOUT;
 	tocentry.cdte_format = CDROM_MSF;
 	if (ioctl (priv->cdrom_fd, CDROMREADTOCENTRY, &tocentry) < 0) {
-		g_warning ("Error getting leadout");
+		g_warning (_("Error getting leadout"));
 		linux_cdrom_invalidate (lcd);
 		return;
 	}
@@ -257,8 +263,8 @@ linux_cdrom_update_cd (LinuxCDRom *lcd)
 	calculate_track_lengths (lcd);
 
 #ifdef DEBUG
-	g_print ("CD changed\n"
-		 "Track count: %d\n------------------\n",
+	g_print (_("CD changed\n"
+		 "Track count: %d\n------------------\n"),
 		 priv->number_tracks);
 #endif
 	
@@ -289,7 +295,7 @@ linux_cdrom_eject (GnomeCDRom *cdrom,
 			if (error) {
 				*error = g_error_new (GNOME_CDROM_ERROR,
 						      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-						      "(eject): ioctl failed: %s",
+						      _("(eject): ioctl failed: %s"),
 						      strerror (errno));
 			}
 
@@ -309,6 +315,8 @@ linux_cdrom_eject (GnomeCDRom *cdrom,
 	}
 
 	g_free (status);
+
+	lcd->priv->ref_count = 0;
 	linux_cdrom_close (lcd);
 	return TRUE;
 }
@@ -332,6 +340,13 @@ linux_cdrom_next (GnomeCDRom *cdrom,
 		return FALSE;
 	}
 
+	if (status->cd != GNOME_CDROM_STATUS_OK) {
+		linux_cdrom_close (lcd);
+		g_free (status);
+		
+		return TRUE;
+	}
+
 	track = status->track + 1;
 	g_free (status);
 	if (track > lcd->priv->number_tracks) {
@@ -380,6 +395,13 @@ linux_cdrom_ffwd (GnomeCDRom *cdrom,
 		return FALSE;
 	}
 
+	if (status->cd != GNOME_CDROM_STATUS_OK) {
+		linux_cdrom_close (lcd);
+		g_free (status);
+		
+		return TRUE;
+	}
+
 	msf = &status->absolute;
 	/* Convert MSF to frames to do calculations on it */
 	frames = msf_to_frames (msf);
@@ -459,7 +481,7 @@ linux_cdrom_play (GnomeCDRom *cdrom,
 			if (error) {
 				*error = g_error_new (GNOME_CDROM_ERROR,
 						      GNOME_CDROM_ERROR_NOT_READY,
-						      "(linux_cdrom_play): Drive not ready");
+						      _("(linux_cdrom_play): Drive not ready"));
 			}
 
 			linux_cdrom_close (lcd);
@@ -479,7 +501,7 @@ linux_cdrom_play (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_NOT_READY,
-					      "(linux_cdrom_play): Drive still not ready");
+					      _("(linux_cdrom_play): Drive still not ready"));
 		}
 
 		linux_cdrom_close (lcd);
@@ -544,7 +566,7 @@ linux_cdrom_play (GnomeCDRom *cdrom,
 			if (error) {
 				*error = g_error_new (GNOME_CDROM_ERROR,
 						      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-						      "(linux_cdrom_play) ioctl failed %s",
+						      _("(linux_cdrom_play) ioctl failed %s"),
 						      strerror (errno));
 			}
 
@@ -580,7 +602,7 @@ linux_cdrom_pause (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_NOT_READY,
-					      "(linux_cdrom_pause): Drive not ready");
+					      _("(linux_cdrom_pause): Drive not ready"));
 		}
 
 		g_free (status);
@@ -593,7 +615,7 @@ linux_cdrom_pause (GnomeCDRom *cdrom,
 			if (error) {
 				*error = g_error_new (GNOME_CDROM_ERROR,
 						      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-						      "(linux_cdrom_pause): Resume failed %s",
+						      _("(linux_cdrom_pause): Resume failed %s"),
 						      strerror (errno));
 			}
 
@@ -612,7 +634,7 @@ linux_cdrom_pause (GnomeCDRom *cdrom,
 			if (error) {
 				*error = g_error_new (GNOME_CDROM_ERROR,
 						      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-						      "(linux_cdrom_pause): ioctl failed %s",
+						      _("(linux_cdrom_pause): ioctl failed %s"),
 						      strerror (errno));
 			}
 
@@ -658,7 +680,7 @@ linux_cdrom_stop (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-					      "(linux_cdrom_stop) ioctl failed %s",
+					      _("(linux_cdrom_stop) ioctl failed %s"),
 					      strerror (errno));
 		}
 
@@ -691,6 +713,13 @@ linux_cdrom_rewind (GnomeCDRom *cdrom,
 		return FALSE;
 	}
 
+	if (status->cd != GNOME_CDROM_STATUS_OK) {
+		linux_cdrom_close (lcd);
+		g_free (status);
+		
+		return TRUE;
+	}
+	
 	msf = &status->absolute;
 
 	frames = msf_to_frames (msf);
@@ -750,6 +779,13 @@ linux_cdrom_back (GnomeCDRom *cdrom,
 		return FALSE;
 	}
 
+	if (status->cd != GNOME_CDROM_STATUS_OK) {
+		linux_cdrom_close (lcd);
+		g_free (status);
+		
+		return TRUE;
+	}
+
 	/* If we're > 0:00 on the track go back to the start of it, 
 	   otherwise go to the previous track */
 	if (status->relative.minute != 0 || status->relative.second != 0) {
@@ -799,8 +835,9 @@ linux_cdrom_get_status (GnomeCDRom *cdrom,
 	LinuxCDRomPrivate *priv;
 	GnomeCDRomStatus *realstatus;
 	struct cdrom_subchnl subchnl;
+	struct cdrom_volctrl vol;
 	int cd_status;
-
+	
 	g_return_val_if_fail (status != NULL, TRUE);
 	
 	lcd = LINUX_CDROM (cdrom);
@@ -808,7 +845,8 @@ linux_cdrom_get_status (GnomeCDRom *cdrom,
 
 	*status = g_new (GnomeCDRomStatus, 1);
 	realstatus = *status;
-
+	realstatus->volume = 0;
+	
 	if (linux_cdrom_open (lcd, error) == FALSE) {
 		g_free (realstatus);
 		linux_cdrom_close (lcd);
@@ -858,7 +896,7 @@ linux_cdrom_get_status (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-					      "(linux_cdrom_get_status): ioctl error %s",
+					      _("(linux_cdrom_get_status): ioctl error %s"),
 					      strerror (errno));
 		}
 
@@ -872,7 +910,7 @@ linux_cdrom_get_status (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-					      "(linux_cdrom_get_status): CDROMSUBCHNL ioctl failed %s",
+					      _("(linux_cdrom_get_status): CDROMSUBCHNL ioctl failed %s"),
 					      strerror (errno));
 		}
 
@@ -881,6 +919,15 @@ linux_cdrom_get_status (GnomeCDRom *cdrom,
 		return FALSE;
 	}
 
+	/* Get the volume */
+	if (ioctl (priv->cdrom_fd, CDROMVOLREAD, &vol) < 0) {
+		g_warning (_("(linux_cdrom_get_status): CDROMVOLREAD ioctl failed %s"),
+			   strerror (errno));
+		realstatus->volume = -1; /* -1 means no volume command */
+	} else {
+		realstatus->volume = vol.channel0;
+	}
+
 	linux_cdrom_close (lcd);
 
 	ASSIGN_MSF (realstatus->relative, blank_msf);
@@ -939,7 +986,7 @@ linux_cdrom_close_tray (GnomeCDRom *cdrom,
 		if (error) {
 			*error = g_error_new (GNOME_CDROM_ERROR,
 					      GNOME_CDROM_ERROR_SYSTEM_ERROR,
-					      "(linux_cdrom_close_tray): ioctl failed %s",
+					      _("(linux_cdrom_close_tray): ioctl failed %s"),
 					      strerror (errno));
 		}
 
@@ -951,6 +998,72 @@ linux_cdrom_close_tray (GnomeCDRom *cdrom,
 	return TRUE;
 }
 
+static gboolean
+linux_cdrom_set_volume (GnomeCDRom *cdrom,
+			int volume,
+			GError **error)
+{
+	LinuxCDRom *lcd;
+	LinuxCDRomPrivate *priv;
+	struct cdrom_volctrl vol;
+
+	lcd = LINUX_CDROM (cdrom);
+	priv = lcd->priv;
+
+	if (linux_cdrom_open (lcd, error) == FALSE) {
+		return FALSE;
+	}
+	
+	vol.channel0 = volume;
+	vol.channel1 = vol.channel2 = vol.channel3 = volume;
+	
+	if (ioctl (priv->cdrom_fd, CDROMVOLCTRL, &vol) < 0) {
+		if (error) {
+			*error = g_error_new (GNOME_CDROM_ERROR,
+					      GNOME_CDROM_ERROR_SYSTEM_ERROR,
+					      _("(linux_cdrom_set_volume:1): ioctl failed %s"),
+					      strerror (errno));
+		}
+
+		linux_cdrom_close (lcd);
+		return FALSE;
+	}
+
+	linux_cdrom_close (lcd);
+	return TRUE;
+}
+
+static gboolean
+linux_cdrom_is_cdrom_device (GnomeCDRom *cdrom,
+			     const char *device,
+			     GError **error)
+{
+	int fd;
+
+	if (device == NULL || *device == 0) {
+		return FALSE;
+	}
+	
+	fd = open (device, O_RDONLY | O_NONBLOCK);
+	if (fd < 0) {
+		return FALSE;
+	}
+
+	/* Fire a harmless ioctl at the device. */
+	if (ioctl (fd, CDROM_GET_CAPABILITY, 0) < 0) {
+		/* Failed, it's not a CDROM drive */
+		g_warning (_("%s is not a CDROM drive"), device);
+		g_print ("Strerror reports: %s\n", strerror (errno));
+		close (fd);
+		
+		return FALSE;
+	}
+	
+	close (fd);
+
+	return TRUE;
+}
+
 static gboolean
 linux_cdrom_get_cddb_data (GnomeCDRom *cdrom,
 			   GnomeCDRomCDDBData **data,
@@ -990,13 +1103,75 @@ linux_cdrom_get_cddb_data (GnomeCDRom *cdrom,
 
 	for (i = priv->track0 - 1; i < priv->track1; i++) {
 		(*data)->offsets[i] = msf_to_frames (&priv->track_info[i].address);
-		g_print ("%d: %u\n", i, msf_to_frames (&priv->track_info[i].address));
 	}
 
 	linux_cdrom_close (lcd);
 	return TRUE;
 }
 
+static gboolean
+linux_cdrom_set_device (GnomeCDRom *cdrom,
+			const char *device,
+			GError **error)
+{
+	LinuxCDRom *lcd;
+	LinuxCDRomPrivate *priv;
+	GnomeCDRomStatus *status;
+	int fd;
+	
+	lcd = LINUX_CDROM (cdrom);
+	priv = lcd->priv;
+
+	if (strcmp (priv->cdrom_device, device) == 0) {
+		/* device is the same */
+		return TRUE;
+	}
+	
+	if (linux_cdrom_get_status (cdrom, &status, NULL) == TRUE) {
+		if (status->audio == GNOME_CDROM_AUDIO_PLAY) {
+			if (linux_cdrom_stop (cdrom, error) == FALSE) {
+				return FALSE;
+			}
+		}
+	}
+
+	if (priv->cdrom_device != NULL) {
+		g_free (priv->cdrom_device);
+	}
+	priv->cdrom_device = g_strdup (device);
+
+	/* Attempt an open to see if we have permission */
+	fd = open (priv->cdrom_device, O_RDONLY | O_NONBLOCK);
+	if (fd < 0 && errno == EACCES && error != NULL) {
+		*error = g_error_new (GNOME_CDROM_ERROR,
+				      GNOME_CDROM_ERROR_NOT_OPENED,
+				      _("You do not seem to have permission to access %s."),
+				      priv->cdrom_device);
+		return FALSE;
+	} else if (gnome_cdrom_is_cdrom_device (GNOME_CDROM (lcd),
+						lcd->priv->cdrom_device, error) == FALSE) {
+		close (fd);
+		if (error) {
+			*error = g_error_new (GNOME_CDROM_ERROR,
+					      GNOME_CDROM_ERROR_NOT_OPENED,
+					      _("%s does not appear to point to a valid CDRom device. This may be because:\n"
+						"a) CD support is not compiled into Linux\n"
+						"b) You do not have the correct permissions to access the CD drive\n"
+						"c) %s is not the CD drive.\n"),					     
+					      lcd->priv->cdrom_device, lcd->priv->cdrom_device);
+		}
+		return FALSE;
+	} else {
+		close (fd);
+	}
+
+	/* Force the new CD to be scanned at next update cycle */
+	g_free (priv->recent_status);
+	priv->recent_status = NULL;
+
+	return TRUE;
+}
+
 static void
 class_init (LinuxCDRomClass *klass)
 {
@@ -1018,10 +1193,14 @@ class_init (LinuxCDRomClass *klass)
 	cdrom_class->back = linux_cdrom_back;
 	cdrom_class->get_status = linux_cdrom_get_status;
 	cdrom_class->close_tray = linux_cdrom_close_tray;
-
+	cdrom_class->set_volume = linux_cdrom_set_volume;
+	cdrom_class->is_cdrom_device = linux_cdrom_is_cdrom_device;
+	
 	/* For CDDB */
   	cdrom_class->get_cddb_data = linux_cdrom_get_cddb_data;
 
+	cdrom_class->set_device = linux_cdrom_set_device;
+	
 	parent_class = g_type_class_peek_parent (klass);
 }
 
@@ -1050,12 +1229,12 @@ update_cd (gpointer data)
 
 	/* Do an update */
 	if (linux_cdrom_get_status (GNOME_CDROM (lcd), &status, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
-
+		gcd_warning ("%s", error);
+		
 		g_error_free (error);
 		return TRUE;
 	}
-
+	
 	if (priv->recent_status == NULL) {
 		priv->recent_status = status;
 		
@@ -1075,7 +1254,7 @@ update_cd (gpointer data)
 			    status->cd == GNOME_CDROM_STATUS_OK) {
 				linux_cdrom_update_cd (lcd);
 			}
-
+			
 			g_free (priv->recent_status);
 			priv->recent_status = status;
 
@@ -1120,38 +1299,54 @@ gnome_cdrom_new (const char *cdrom_device,
 	LinuxCDRomPrivate *priv;
 	int fd;
 
-	g_return_val_if_fail (cdrom_device != NULL, NULL);
-	g_return_val_if_fail (*cdrom_device != 0, NULL);
-
 	cdrom = g_object_new (linux_cdrom_get_type (), NULL);
 	priv = cdrom->priv;
 
 	priv->cdrom_device = g_strdup (cdrom_device);
 	priv->update = update;
 
-	/* Do a test open to see if we have CD stuff working */
-	if (linux_cdrom_open (cdrom, error) == TRUE) {
-
-		/* Force an update so that a status will always exist */
-		update_cd (cdrom);
-		
-		/* All worked, start the counter */
-		switch (update) {
-		case GNOME_CDROM_UPDATE_NEVER:
-			break;
-			
-		case GNOME_CDROM_UPDATE_WHEN_CHANGED:
-		case GNOME_CDROM_UPDATE_CONTINOUS:
-			priv->update_id = g_timeout_add (1000, update_cd, cdrom);
-			break;
-			
-		default:
-			break;
+	fd = open (cdrom_device, O_RDONLY | O_NONBLOCK);
+	
+	if (fd < 0) {
+		if (errno == EACCES &&
+		    error != NULL) {
+			*error = g_error_new (GNOME_CDROM_ERROR,
+					      GNOME_CDROM_ERROR_NOT_OPENED,
+					      _("You do not seem to have permission to access %s."),
+					      cdrom_device);
+		}
+	} else if (gnome_cdrom_is_cdrom_device (GNOME_CDROM (cdrom),
+						cdrom->priv->cdrom_device,
+						error) == FALSE) {
+		close (fd);
+		if (error) {
+			*error = g_error_new (GNOME_CDROM_ERROR,
+					      GNOME_CDROM_ERROR_NOT_OPENED,
+					      _("%s does not appear to point to a valid CDRom device. This may be because:\n"
+						"a) CD support is not compiled into Linux\n"
+						"b) You do not have the correct permissions to access the CD drive\n"
+						"c) %s is not the CD drive.\n"),					     
+					      cdrom->priv->cdrom_device, cdrom->priv->cdrom_device);
 		}
-		
-		linux_cdrom_close (cdrom);
 	} else {
-		return NULL;
+		close (fd);
+	}
+
+	/* Force an update so that a status will always exist */
+	update_cd (cdrom);
+	
+	/* All worked, start the counter */
+	switch (update) {
+	case GNOME_CDROM_UPDATE_NEVER:
+		break;
+		
+	case GNOME_CDROM_UPDATE_WHEN_CHANGED:
+	case GNOME_CDROM_UPDATE_CONTINOUS:
+		priv->update_id = g_timeout_add (1000, update_cd, cdrom);
+		break;
+		
+	default:
+		break;
 	}
 
 	return GNOME_CDROM (cdrom);
