@@ -27,7 +27,10 @@
 #include <sys/types.h>
 #include <dirent.h>
 
+#include <atk/atk.h>
+
 #include <gtk/gtkdialog.h>
+#include <gtk/gtkmessagedialog.h>
 #include <gtk/gtkstock.h>
 #include <gtk/gtkvbox.h>
 #include <gtk/gtkhbox.h>
@@ -56,13 +59,29 @@ static void
 do_device_changed (GnomeCDPreferences *prefs,
 		   const char *device)
 {
+	GError *error = NULL;
+	gboolean ret;
+	
 	if (prefs->device != NULL) {
 		g_free (prefs->device);
 	}
 
 	prefs->device = g_strdup (device);
 	if (prefs->gcd->cdrom != NULL) {
-		gnome_cdrom_set_device (prefs->gcd->cdrom, prefs->device, NULL);
+		ret = gnome_cdrom_set_device (prefs->gcd->cdrom, prefs->device, &error);
+		if (ret == FALSE) {
+			GtkWidget *dialog;
+
+			dialog = gtk_message_dialog_new (NULL, 0,
+							 GTK_MESSAGE_ERROR,
+							 GTK_BUTTONS_OK,
+							 _("%s\nThis means that Gnome-CD will not be able to run."), error->message);
+			gtk_window_set_title (GTK_WINDOW (dialog), _("Error setting device"));
+			g_signal_connect (G_OBJECT (dialog), "response",
+					  G_CALLBACK (gtk_widget_destroy), dialog);
+			gtk_widget_show (dialog);
+			g_error_free (error);
+		}
 	}
 }
 
@@ -92,6 +111,7 @@ on_start_changed (GConfClient *_client,
 	   so we don't need to do anything for it */
 }
 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 static void
 close_on_start_changed (GConfClient *_client,
 			guint cnxn_id,
@@ -104,6 +124,7 @@ close_on_start_changed (GConfClient *_client,
 	prefs->start_close = gconf_value_get_bool (value);
 	/* This doesn't take effect till next start either */
 }
+#endif
 
 static void
 on_stop_changed (GConfClient *_client,
@@ -122,12 +143,23 @@ static void
 do_theme_changed (GnomeCDPreferences *prefs,
 		  const char *theme_name)
 {
+	GCDTheme *old_theme;
+	
 	if (prefs->theme_name != NULL) {
 		g_free (prefs->theme_name);
 	}
 
 	prefs->theme_name = g_strdup (theme_name);
-	/* FIXME: Change */
+	old_theme = prefs->gcd->theme;
+	prefs->gcd->theme = theme_load (prefs->gcd, theme_name);
+
+	/* Revert to the old theme if something messed up */
+	if (prefs->gcd->theme == NULL) {
+		prefs->gcd->theme = old_theme;
+	} else {
+		theme_change_widgets (prefs->gcd);
+		theme_free (old_theme);
+	}
 }
 
 static void
@@ -157,6 +189,11 @@ restore_preferences (GnomeCDPreferences *prefs)
 	
 	prefs->device = gconf_client_get_string (client,
 						 "/apps/gnome-cd/device", NULL);
+	if (prefs->device == NULL) {
+		g_warning ("GConf schemas are not correctly installed.");
+		prefs->device = g_strdup ("/dev/cdrom");
+	}
+	
 	prefs->device_id = gconf_client_notify_add (client,
 						    "/apps/gnome-cd/device", device_changed,
 						    prefs, NULL, &error);
@@ -170,13 +207,14 @@ restore_preferences (GnomeCDPreferences *prefs)
 						   "/apps/gnome-cd/on-start",
 						   on_start_changed, prefs,
 						   NULL, NULL);
-	
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL	
 	prefs->start_close = gconf_client_get_bool (client,
 						    "/apps/gnome-cd/close-on-start", NULL);
 	prefs->close_id = gconf_client_notify_add (client,
 						   "/apps/gnome-cd/close-on-start",
 						   close_on_start_changed, prefs,
 						   NULL, NULL);
+#endif
 	
 	prefs->stop = gconf_client_get_int (client,
 					    "/apps/gnome-cd/on-stop", NULL);
@@ -187,6 +225,11 @@ restore_preferences (GnomeCDPreferences *prefs)
 
 	prefs->theme_name = gconf_client_get_string (client,
 						     "/apps/gnome-cd/theme-name", NULL);
+	if (prefs->theme_name == NULL) {
+		g_warning ("GConf schemas are not correctly installed");
+		prefs->theme_name = g_strdup ("lcd");
+	}
+	
 	prefs->theme_id = gconf_client_notify_add (client,
 						   "/apps/gnome-cd/theme-name",
 						   theme_changed, prefs,
@@ -209,7 +252,9 @@ preferences_free (GnomeCDPreferences *prefs)
 	/* Remove the listeners */
 	gconf_client_notify_remove (client, prefs->device_id);
 	gconf_client_notify_remove (client, prefs->start_id);
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	gconf_client_notify_remove (client, prefs->close_id);
+#endif
 	gconf_client_notify_remove (client, prefs->stop_id);
 	gconf_client_notify_remove (client, prefs->theme_id);
 
@@ -241,18 +286,28 @@ typedef struct _PropertyDialog {
 	GtkWidget *start_nothing;
 	GtkWidget *start_play;
 	GtkWidget *start_stop;
+	
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	GtkWidget *start_close;
+#endif
 
 	GtkWidget *stop_nothing;
 	GtkWidget *stop_stop;
 	GtkWidget *stop_open;
+	
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	GtkWidget *stop_close;
+#endif
 
 	GtkWidget *theme_list;
 	
 	guint start_id;
 	guint device_id;
+	
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	guint close_id;
+#endif
+	
 	guint stop_id;
 	guint theme_id;
 } PropertyDialog;
@@ -283,7 +338,9 @@ prefs_destroy_cb (GtkDialog *dialog,
 	gconf_client_notify_remove (client, pd->device_id);
 	gconf_client_notify_remove (client, pd->start_id);
 	gconf_client_notify_remove (client, pd->stop_id); 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	gconf_client_notify_remove (client, pd->close_id);
+#endif
 	gconf_client_notify_remove (client, pd->theme_id);
 	
 	g_free (pd);
@@ -296,13 +353,17 @@ apply_clicked_cb (GtkWidget *apply,
 	char *new_device;
 
 	new_device = gtk_entry_get_text (GTK_ENTRY (pd->cd_device));
-	if (new_device == NULL || pd->gcd->preferences->device == NULL) {
+	
+	if (new_device == NULL) {
 		return;
 	}
+
+	if (pd->gcd->preferences->device != NULL) {
+		if (strcmp (new_device, pd->gcd->preferences->device) == 0) {
+			return;
+		}
+	}
 	
-	if (strcmp (new_device, pd->gcd->preferences->device) == 0) {
-		return;
-	}
 	gconf_client_set_string (client, "/apps/gnome-cd/device", new_device, NULL);
 	gtk_widget_set_sensitive (pd->apply, FALSE);
 }
@@ -429,6 +490,7 @@ stop_open_toggled_cb (GtkToggleButton *tb,
 	set_stop (pd, GNOME_CD_PREFERENCES_STOP_OPEN);
 }
 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 static void
 stop_close_toggled_cb (GtkToggleButton *tb,
 		       PropertyDialog *pd)
@@ -439,6 +501,7 @@ stop_close_toggled_cb (GtkToggleButton *tb,
 
 	set_stop (pd, GNOME_CD_PREFERENCES_STOP_CLOSE);
 }
+#endif
 
 static void
 change_start_widget (GConfClient *client,
@@ -505,12 +568,14 @@ change_stop_widget (GConfClient *client,
 		tb = GTK_TOGGLE_BUTTON (pd->stop_open);
 		func = G_CALLBACK (stop_open_toggled_cb);
 		break;
-
+		
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	case GNOME_CD_PREFERENCES_STOP_CLOSE:
 		tb = GTK_TOGGLE_BUTTON (pd->stop_close);
 		func = G_CALLBACK (stop_close_toggled_cb);
 		break;
-
+#endif
+		
 	default:
 		g_warning ("Unknown stop value: %d", gconf_value_get_int (value));
 		return;
@@ -523,6 +588,7 @@ change_stop_widget (GConfClient *client,
 					   0, 0, NULL, func, pd);
 }
 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 static void
 start_close_toggled_cb (GtkToggleButton *tb,
 			PropertyDialog *pd)
@@ -550,6 +616,7 @@ change_start_close_widget (GConfClient *client,
 	g_signal_handlers_unblock_matched (G_OBJECT (pd->start_close), G_SIGNAL_MATCH_FUNC,
 					   0, 0, NULL, G_CALLBACK (start_close_toggled_cb), pd);
 }
+#endif
 
 static void
 theme_selection_changed_cb (GtkTreeSelection *selection,
@@ -558,12 +625,10 @@ theme_selection_changed_cb (GtkTreeSelection *selection,
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 
-	g_print ("Hello?\n");
 	if (gtk_tree_selection_get_selected (selection, &model, &iter) == TRUE) {
 		char *theme_name;
 
 		gtk_tree_model_get (model, &iter, 0, &theme_name, -1);
-		g_print ("Boink!\n");
 		gconf_client_set_string (client, "/apps/gnome-cd/theme-name", theme_name, NULL);
 	}
 }
@@ -641,7 +706,51 @@ create_theme_model (PropertyDialog *pd,
 
 	return GTK_TREE_MODEL (store);
 }
-	
+
+static void
+add_relation (AtkRelationSet *set,
+	      AtkRelationType type,
+	      AtkObject *target)
+{
+	AtkRelation *relation;
+
+	relation = atk_relation_set_get_relation_by_type (set, type);
+
+	if (relation != NULL) {
+		GPtrArray *array = atk_relation_get_target (relation);
+
+		g_ptr_array_remove (array, target);
+		g_ptr_array_add (array, target);
+	} else {
+		/* Relation hasn't been created yet */
+		relation = atk_relation_new (&target, 1, type);
+
+		atk_relation_set_add (set, relation);
+		g_object_unref (relation);
+	}
+}
+
+static void
+add_paired_relations (GtkWidget *target1,
+		      AtkRelationType target1_type,
+		      GtkWidget *target2,
+		      AtkRelationType target2_type)
+{
+	AtkObject *atk_target1;
+	AtkObject *atk_target2;
+	AtkRelationSet *set1;
+	AtkRelationSet *set2;
+
+	atk_target1 = gtk_widget_get_accessible (target1);
+	atk_target2 = gtk_widget_get_accessible (target2);
+
+	set1 = atk_object_ref_relation_set (atk_target1);
+	add_relation (set1, target1_type, atk_target2);
+
+	set2 = atk_object_ref_relation_set (atk_target2);
+	add_relation (set2, target2_type, atk_target1);
+}
+
 GtkWidget *
 preferences_dialog_show (GnomeCD *gcd,
 			 gboolean only_device)
@@ -689,6 +798,10 @@ preferences_dialog_show (GnomeCD *gcd,
 
 	pd->cd_device = gtk_entry_new ();
 	gtk_entry_set_text (GTK_ENTRY (pd->cd_device), gcd->preferences->device);
+	gtk_label_set_mnemonic_widget (GTK_LABEL (label), pd->cd_device);
+	add_paired_relations (label, ATK_RELATION_LABEL_FOR,
+			      pd->cd_device, ATK_RELATION_LABELLED_BY);
+	
 	g_signal_connect (G_OBJECT (pd->cd_device), "changed",
 			  G_CALLBACK (device_changed_cb), pd);
 	g_signal_connect (G_OBJECT (pd->cd_device), "activate",
@@ -746,6 +859,7 @@ preferences_dialog_show (GnomeCD *gcd,
 						"/apps/gnome-cd/on-start",
 						change_start_widget, pd, NULL, NULL);
 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL	
 	pd->start_close = gtk_check_button_new_with_mnemonic (_("Attempt to _close CD tray"));
 	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (pd->start_close),
 				      gcd->preferences->start_close);
@@ -756,6 +870,7 @@ preferences_dialog_show (GnomeCD *gcd,
 	pd->close_id = gconf_client_notify_add (client,
 						"/apps/gnome-cd/close-on-start",
 						change_start_close_widget, pd, NULL, NULL);
+#endif
 	/* Right side */
 	frame = gtk_frame_new (_("When Gnome-CD quits"));
 	gtk_container_set_border_width (GTK_CONTAINER (frame), 2);
@@ -791,6 +906,7 @@ preferences_dialog_show (GnomeCD *gcd,
 			  G_CALLBACK (stop_open_toggled_cb), pd);
 	gtk_box_pack_start (GTK_BOX (vbox), pd->stop_open, FALSE, FALSE, 0);
 
+#ifdef HAVE_CDROMCLOSETRAY_IOCTL
 	pd->stop_close = gtk_radio_button_new_with_mnemonic_from_widget (GTK_RADIO_BUTTON (pd->stop_open),
 									 _("Attempt to c_lose CD tray"));
 	if (gcd->preferences->stop == GNOME_CD_PREFERENCES_STOP_CLOSE) {
@@ -799,11 +915,12 @@ preferences_dialog_show (GnomeCD *gcd,
 	g_signal_connect (G_OBJECT (pd->stop_close), "toggled",
 			  G_CALLBACK (stop_close_toggled_cb), pd);
 	gtk_box_pack_start (GTK_BOX (vbox), pd->stop_close, FALSE, FALSE, 0);
-
+#endif
+	
 	pd->stop_id = gconf_client_notify_add (client,
 					       "/apps/gnome-cd/on-stop",
 					       change_stop_widget, pd, NULL, NULL);
-
+	
 	if (only_device == TRUE) {
 		gtk_widget_set_sensitive (hbox, FALSE);
 	}
