@@ -165,12 +165,12 @@ main (int argc,
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 	textdomain (GETTEXT_PACKAGE);
 	
-	program = gnome_program_init ("gnome-sound-recorder", VERSION,
+	program = gnome_program_init ("grecord", VERSION,
 				      LIBGNOMEUI_MODULE, argc, argv,
 				      GNOME_PARAM_POPT_TABLE, grec_options,
 				      GNOME_PARAM_HUMAN_READABLE_NAME,
 				      _("Sound recorder"),
-				      NULL);
+				      GNOME_PARAM_APP_DATADIR,DATADIR,NULL);
 	
 	gnome_window_icon_set_default_from_file (GNOME_ICONDIR"/gnome-audio2.png");
 	g_value_init (&value, G_TYPE_POINTER);
@@ -243,6 +243,7 @@ main (int argc,
 	p = g_find_program_in_path (sox_command);
 	if (p == NULL) {
 		able_to_record = FALSE;
+		gtk_widget_set_sensitive (GTK_WIDGET (grecord_widgets.Record_button), FALSE);
 		if (show_warningmess) {
 			GtkWidget* dont_show_again_checkbutton = gtk_check_button_new_with_label (_("Don't show this message again."));
 			
